/*
Copyright (C) 2014 Paul Boddie

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
*/

module cartridge()
{
	$fn = 50;

	/* Configure the generated shapes. */

	BACK_CAVITY = 1;
	FRONT_LABEL_INSET = 1;
	TOP_LABEL_INSET = 1;
	GROOVE = 1;

	BACK = 1; BACK_SURFACE = 1;
	FRONT = 1; FRONT_SURFACE = 1;
	APART = 1;

	/*
	Rounding/fillet radius and additional margin of subtracted
	material. The additional margin helps avoid geometry problems.
	*/

	rr = 2;
	ro = rr;
	extra = 0.1;

	module fillet(r, h) {
		translate([0, 0, -h/2])
			difference() {
				cube([r + extra, r + extra, h + extra]);
				cylinder(r = r, h = h);
			}
	}

	/*
	Justify an object of the given dimensions, according to the given
	factors (where 1 indicates moving the object to the positive side of an
	axis, and -1 indicates moving it to the negative side of an axis).

	NOTE: child should eventually be replaced by children.
	*/
	module justify(width, depth, height, wdir, ddir, hdir) {
		translate([
			wdir * width / 2,
			ddir * depth / 2,
			hdir * height / 2])
			child();
	}

	/*
	Make a cuboid of the given dimensions, justifying it according to the given
	factors, and moving it to the specified coordinates.

	NOTE: Usage of justify within this module will not work due to recursion
	NOTE: limitations in openscad, potentially removed in more recent
	NOTE: releases. Thus, the justify transform is merged in here.
	*/
	module cube_at(width, depth, height, wdir, ddir, hdir, x, y, z) {
		translate([
			x + wdir * width / 2,
			y + ddir * depth / 2,
			z + hdir * height / 2])
			cube([width, depth, height], center = true);
	}

	/* Internal dimensions. */

	int_width = 86.0;
	int_depth = 11.0;
	int_payload_height = 50.8;	/* space between the top and the floor */
	int_connector_height = 13.5; 	/* vertical offset of bottom/floor of payload area */

	/* Side thicknesses. */

	front = 2;
	back = 3.5;
	top = 3;
	side = 2; 			/* increased from 1.5 for 3D printing reliability */
	bottom = 1;			/* thickness of floor of payload area */

	/* How much extra depth the back provides for mating with the front. */

	groove_width_extra = 1.0;
	front_back_overlap = 1.0;
	groove_width_overlap = front_back_overlap + groove_width_extra;

	/* Division of pieces into front and back, defining the extents. */

	int_front_depth = 4.5;
	front_depth = int_front_depth + front;
	int_back_depth = int_depth - int_front_depth + front_back_overlap;
	back_depth = int_back_depth + back;

	/* Cartridge dimensions. */

	width = int_width + side + side;
	height = top + int_payload_height + bottom + int_connector_height;
	depth = int_depth + front + back;
	int_payload_upper_extent = height / 2 - top;

	/* Cartridge surfaces. */

	front_side = side;
	front_left = front_side;
	front_right = front_side;
	back_side = side;
	back_left = back_side;
	back_right = back_side;

	/* Label details. */

	front_label_width = 83.0;
	front_label_height = 46.0;
	front_label_depth = 1.0;
	front_label_offset_from_bottom = 19.5;

	top_label_width = front_label_width;
	top_label_height = 11.5; /* the height as seen from above */
	top_label_depth = front_label_depth;
	top_label_offset_from_front = 2.5;

	/*
	The groove around the sides and top. This is extended to allow the
	pieces to fit together, and this extension is generated regardless of
	whether the visible groove is enabled or not.
	*/

	groove_width_exposed = GROOVE ? 1.5 : 0;
	groove_width = groove_width_exposed + groove_width_overlap;
	groove_depth = 1.0; /* how deep the groove goes into each side */

	/* Additional cutting to mate the back and front. */

	top_groove_width = groove_width_overlap;
	top_groove_depth = 2.0;

	/*
	Space for the inner edge of the back inside the front.
	Offsets are measured from the outside surfaces.
	*/

	inner_top_front_cutout_width = int_width;
	inner_top_front_cutout_depth = top_groove_width;
	inner_top_front_cutout_height = top - top_groove_depth;

	inner_side_front_cutout_height = height - top_groove_depth;
	inner_side_front_cutout_width = front_side - groove_depth;
	inner_side_front_cutout_depth = groove_width_overlap;

	/*
	The back cavity is the indented part at the bottom of the back of the
	cartridge. It appears to be mechanically superfluous, just making the
	shape of the cartridge look a bit fancier.
	*/

	back_cavity_width = 68.0;
	back_cavity_inner_width = 65.0;
	back_cavity_offset_from_inner_left = 9.0;
	back_cavity_inner_offset_from_inner_left = 10.5;
	back_cavity_height = 13.5;
	back_cavity_inner_height = 12.0;
	back_cavity_depth = 1.5;

	/*
	The effect of the cavity on the inside of the case. The most important
	measurement is the maximum offset since it defines the width of the
	internal space that should accommodate the plastic guides of the Plus 1
	socket.
	*/

	inner_back_slope_depth = 2.5;
	inner_back_slope_width = inner_back_slope_depth;
	inner_back_slope_max_offset = 10.5;
	inner_back_slope_min_offset = inner_back_slope_max_offset - inner_back_slope_width;

	/* The tapering off of the inner back edge. */

	inner_back_edge_width = 69.0;
	inner_back_edge_height = 3.0;
	inner_back_edge_depth = 1.5;

	/* The tapering off of the inner front edge. */

	inner_front_edge_width = width - front_side * 2;
	inner_front_edge_height = 3.0;
	inner_front_edge_depth = 1.5;

	/*
	The cutouts in the floor of the front of the cartridge that produce a
	kind of tab that guides the edge connector into place in the back cutout.
	*/

	edge_connector_cutout_front_offset = 1.0;
	edge_connector_cutout_front_depth = front_back_overlap;
	edge_connector_cutout_front_width = 15.0;

	/*
	The cutout in the floor of the back of the cartridge that accommodates
	the edge connector.
	*/

	edge_connector_cutout_back_depth = 3.0;
	edge_connector_cutout_back_width = 57.5;

	/*
	Edge connectors are themselves 0.05" or approximately 1.27mm in
	thickness according to the Acorn Electron Cartridge Interface Specification
	(Acorn Support Application Group Note 014).
	*/

	/* Extra internal features. Depths include front and back surfaces. */

	pcb_back_support_width = 1.2;
	pcb_back_support_depth = back_depth -
		edge_connector_cutout_back_depth;
	pcb_back_support_height = height - int_connector_height - top - bottom;

	pcb_front_support_width = 1.2;
	pcb_front_support_depth = front_depth;
	pcb_front_support_height = pcb_back_support_height;

	/*
	Features measured from the Stardot Dual ROM Adaptor	cartridge board
	dimensions diagram. Offsets are from inside the bottom "floor".
	*/

	pcb_back_support_bump_width = pcb_front_support_width;
	pcb_back_support_bump_depth = 1.5;
	pcb_back_support_left_bump_height = 13.2;
	pcb_back_support_right_bump_height = 10.7;
	pcb_back_support_left_bump_offset_from_bottom = 15.1;
	pcb_back_support_right_bump_offset_from_bottom = 17.6;
	pcb_back_support_top_bump_height = 3.8;

	/* Move the PCB support towards the centre. */
	pcb_support_margin = 1.75;
	pcb_support_offset_from_centre = edge_connector_cutout_back_width / 2
		- pcb_support_margin;

	pcb_lug_depth = pcb_back_support_depth +
		pcb_back_support_bump_depth;
	pcb_lug_inner_radius = 1.0;
	pcb_lug_outer_radius = 5.5 / 2;
	pcb_lug_offset_from_bottom = 14.35;
	pcb_lug_offset_from_outside = 7.55;

	pcb_lug_cross_width = 6.7;
	pcb_lug_cross_depth = pcb_back_support_depth;
	pcb_lug_cross_height = 1.4;

	pcb_front_lug_depth = pcb_back_support_bump_depth +
		pcb_front_support_depth - front;
	pcb_front_lug_inner_radius = pcb_lug_outer_radius;
	pcb_front_lug_outer_radius = pow(
		pow(pcb_lug_cross_width / 2, 2) +
		pow(pcb_lug_cross_height / 2, 2),
		0.5);

	/* Repeated constructs. */

	module pcb_support(xdir, bump_height, bump_offset) {

		/*
		Translate the support in the stated direction.
		Since the support is already justified in the direction, the translation
		moves the inner edge of the support to alignment with the end of the
		edge connector cutout and then back by the PCB support margin.
		*/

		translate([xdir *
			pcb_support_offset_from_centre,
			edge_connector_cutout_back_depth,
			-height / 2 + bottom + int_connector_height])
			justify(pcb_back_support_width,
				pcb_back_support_depth,
				pcb_back_support_height,
				xdir, 1, 1)
				union() {

					/* Underlying support strut. */

					cube([pcb_back_support_width,
						pcb_back_support_depth,
						pcb_back_support_height], center = true);

					/* Middle bump. */

					cube_at(pcb_back_support_bump_width,
						pcb_back_support_bump_depth,
						bump_height,
						0, -1, 1,
						0,
						-pcb_back_support_depth / 2,
						-pcb_back_support_height / 2 + bump_offset);

					/* Top bump. */

					cube_at(pcb_back_support_bump_width,
						pcb_back_support_bump_depth,
						pcb_back_support_top_bump_height,
						0, -1, 1,
						0,
						-pcb_back_support_depth / 2,
						pcb_back_support_height / 2 -
							pcb_back_support_top_bump_height);
				}
	}

	module pcb_lug(xdir) {
		translate([xdir *
			(width/2 - pcb_lug_offset_from_outside),
			back_depth,
			-height / 2 + bottom + int_connector_height +
				pcb_lug_offset_from_bottom
			])
			rotate([90, 0, 0])
				difference() {
					union() {
						cylinder(h=pcb_lug_depth, r=pcb_lug_outer_radius);
						cube_at(pcb_lug_cross_width,
							pcb_lug_cross_height, pcb_lug_cross_depth,
							0, 0, 1,
							0, 0, 0);
						cube_at(pcb_lug_cross_height,
							pcb_lug_cross_width, pcb_lug_cross_depth,
							0, 0, 1,
							0, 0, 0);
					}
					cylinder(h=pcb_lug_depth, r=pcb_lug_inner_radius);
				}
	}

	module pcb_front_lug(xdir) {
		translate([xdir *
			(width/2 - pcb_lug_offset_from_outside),
			-int_front_depth + pcb_front_lug_depth,
			-height / 2 + bottom + int_connector_height +
				pcb_lug_offset_from_bottom
			])
			rotate([90, 0, 0])
				difference() {
					cylinder(h=pcb_front_lug_depth,
						r=pcb_front_lug_outer_radius);
					cylinder(h=pcb_front_lug_depth,
						r=pcb_front_lug_inner_radius);
				}
	}

	/* The actual shapes. */

	front_displacement = APART ? -width * 0.6 : 0;

	if (FRONT)
		translate([front_displacement, 0, 0])
		difference() {

			/* The cartridge surfaces. */

			union() {

				/* Front portion. */

				if (FRONT_SURFACE) {
					cube_at(width, front, height,
						0, -1, 0,
						0, -int_front_depth, 0);
					cube_at(front_left, front_depth, height,
						-1, -1, 0,
						-int_width / 2, 0, 0);
					cube_at(front_right, front_depth, height,
						1, -1, 0,
						int_width / 2, 0, 0);
					cube_at(width, front_depth, top,
						0, -1, 1,
						0, 0, int_payload_upper_extent);
				}

				difference() {

					/* Floor of cartridge. */

					cube_at(width, front_depth, bottom,
						0, -1, 1,
						0, 0, -height / 2 + int_connector_height);

					/* Left cutout. */

					cube_at(edge_connector_cutout_front_width,
						edge_connector_cutout_front_depth,
						bottom,
						1, -1, 1,
						-width / 2 + edge_connector_cutout_front_offset,
						0,
						-height / 2 + int_connector_height);

					/* Right cutout. */

					cube_at(edge_connector_cutout_front_width,
						edge_connector_cutout_front_depth,
						bottom,
						-1, -1, 1,
						width / 2 - edge_connector_cutout_front_offset,
						0,
						-height / 2 + int_connector_height);
				}

				/* PCB supports. */

				cube_at(pcb_front_support_width,
						pcb_front_support_depth,
						pcb_front_support_height,
						1, -1, 1,
						-edge_connector_cutout_back_width / 2 +
							pcb_support_margin,
						0,
						-height / 2 + bottom + int_connector_height);

				cube_at(pcb_front_support_width,
						pcb_front_support_depth,
						pcb_front_support_height,
						-1, -1, 1,
						edge_connector_cutout_back_width / 2 -
							pcb_support_margin,
						0,
						-height / 2 + bottom + int_connector_height);

				/* Circular "lugs" to hold PCBs in place. */

				pcb_front_lug(-1);
				pcb_front_lug(1);
			}

			/* Label insets. */

			union() {

				/* Front label. */

				if (FRONT_LABEL_INSET)
					translate([-front_label_width / 2, -front_depth,
						front_label_offset_from_bottom - height / 2])
						cube([front_label_width, front_label_depth,
							front_label_height]);

				/* Top label. See also the back piece. */

				if (TOP_LABEL_INSET)
					translate([-top_label_width / 2,
						-front_depth + top_label_offset_from_front,
						height / 2 - top_label_depth])
						cube([top_label_width, top_label_height,
							top_label_depth]);
			}

			/* Inner front edge cavity. */

			translate([inner_front_edge_width / 2, -int_front_depth, -height / 2])
				rotate([0, -90, 0])
					linear_extrude(height = inner_front_edge_width)
						polygon([
							[-extra, -inner_front_edge_depth],
							[0, -inner_front_edge_depth],
							[inner_front_edge_height, 0],
							[-extra, 0],
							]);

			/* Inner top cutout for the top and sides of the back portion. */

			cube_at(inner_top_front_cutout_width,
				inner_top_front_cutout_depth,
				inner_top_front_cutout_height,
				0, -1, 1,
				0, 0, int_payload_upper_extent);

			cube_at(inner_side_front_cutout_width,
				inner_side_front_cutout_depth,
				inner_side_front_cutout_height,
				1, -1, 0,
				int_width / 2, 0, -inner_top_front_cutout_height);

			cube_at(inner_side_front_cutout_width,
				inner_side_front_cutout_depth,
				inner_side_front_cutout_height,
				-1, -1, 0,
				-int_width / 2, 0, -inner_top_front_cutout_height);

			/* Fillets to round off the edges. */

			union() {

				/* Top left and right rounding. */
			
				translate([-width / 2 + ro, -front_depth / 2, height / 2 - ro])
					rotate([0, 0, 180])
						rotate([90, 0, 0])
							fillet(rr, front_depth);
				translate([width / 2 - ro, -front_depth / 2, height / 2 - ro])
					rotate([90, 0, 0])
						fillet(rr, front_depth);

				/* Top front rounding. */
			
				translate([0, -front_depth + ro, height / 2 - ro])
					rotate([0, 0, 180])
						rotate([0, -90, 0])
							fillet(rr, width);

				/* Edge rounding. */
			
				translate([-width / 2 + ro, -front_depth + ro, 0])
					rotate([0, 0, 180])
						fillet(rr, height);
				translate([width / 2 - ro, -front_depth + ro, 0])
					rotate([0, 0, 270])
						fillet(rr, height);
			}
		}

	/*
	Place the back piece next to the front, with the internals facing out
	the same way, if APART is defined.
	*/

	back_displacement = APART ? width * 0.6 : 0;
	back_rotation = APART ? 180 : 0;

	if (BACK)
		translate([back_displacement, 0, 0])
		rotate([0, 0, back_rotation])
		difference() {

			/* The cartridge surfaces. */

			union() {

				/* Back portion. */

				if (BACK_SURFACE) {
					cube_at(width, back, height,
						0, 1, 0,
						0, int_back_depth, 0);
					cube_at(back_left, back_depth + groove_width_extra, height,
						-1, 1, 0,
						-int_width / 2, -groove_width_extra, 0);
					cube_at(back_right, back_depth + groove_width_extra, height,
						1, 1, 0,
						int_width / 2, -groove_width_extra, 0);
					cube_at(width, back_depth + groove_width_extra, top,
						0, 1, 1,
						0, -groove_width_extra, int_payload_upper_extent);
				}

				difference() {

					/* Floor of cartridge. */

					cube_at(width, back_depth, bottom,
						0, 1, 1,
						0, 0, -height / 2 + int_connector_height);

					/* Edge connector cutout. */

					cube_at(edge_connector_cutout_back_width,
						edge_connector_cutout_back_depth,
						bottom,
						0, 1, 1,
						0, 0, -height / 2 + int_connector_height);
				}

				/* PCB supports. */

				pcb_support(-1, pcb_back_support_left_bump_height,
					pcb_back_support_left_bump_offset_from_bottom);
				pcb_support(1, pcb_back_support_right_bump_height,
					pcb_back_support_right_bump_offset_from_bottom);

				/* Circular "lugs" to hold PCBs in place. */

				pcb_lug(-1);
				pcb_lug(1);
			}

			/* Label insets. */

			union() {

				/* Top label. See also the front piece. */

				if (TOP_LABEL_INSET)
					translate([-top_label_width / 2,
						-front_depth + top_label_offset_from_front,
						height / 2 - top_label_depth])
						cube([top_label_width, top_label_height,
							top_label_depth]);
			}

			/* Top and side grooves, positioned in the back portion. */

			union() {

				/* Left groove. */

				cube_at(groove_depth, groove_width, height,
					1, 1, 0,
					-width / 2, -groove_width_extra, 0);

				/* Right groove. */

				cube_at(groove_depth, groove_width, height,
					-1, 1, 0,
					width / 2, -groove_width_extra, 0);

				/* Top grooves. */

				cube_at(width, groove_width, groove_depth,
					0, 1, -1,
					0, -groove_width_extra, height / 2);

				cube_at(width, top_groove_width, top_groove_depth,
					0, 1, -1,
					0, -groove_width_extra, height / 2);
			}

			/* Back cavity. */

			if (BACK_CAVITY)
				intersection() {

					/* From the bottom upwards. */

					translate([0, back_depth, -height / 2])
						linear_extrude(height = back_cavity_height)
							translate([-int_width / 2, 0, 0])
								polygon([
									[back_cavity_offset_from_inner_left, 0],
									[back_cavity_inner_offset_from_inner_left,
										-back_cavity_depth],
									[back_cavity_inner_offset_from_inner_left +
										back_cavity_inner_width,
										-back_cavity_depth],
									[back_cavity_offset_from_inner_left +
										back_cavity_width, 0]
									]);

					/* From left to right. */

					translate([back_cavity_width / 2, back_depth, -height / 2])
						rotate([0, -90, 0])
							linear_extrude(height = back_cavity_width)
								polygon([
									[-extra, -back_cavity_depth],
									[back_cavity_inner_height,
										-back_cavity_depth],
									[back_cavity_height, 0],
									[-extra, 0]
									]);
				}

			/* Inner back cavities. */

			translate([0, int_back_depth, -height / 2])
				linear_extrude(height = int_connector_height)
					translate([-int_width / 2, 0, 0])
						polygon([
							[0, 0],
							[inner_back_slope_max_offset, 0],
							[inner_back_slope_min_offset,
								inner_back_slope_depth],
							[0, inner_back_slope_depth]
							]);

			translate([0, int_back_depth, -height / 2])
				linear_extrude(height = int_connector_height)
					translate([int_width / 2, 0, 0])
						polygon([
							[0, 0],
							[-inner_back_slope_max_offset, 0],
							[-inner_back_slope_min_offset,
								inner_back_slope_depth],
							[0, inner_back_slope_depth]
							]);

			/* Inner back edge cavity. */

			translate([inner_back_edge_width / 2,
				int_back_depth + inner_back_edge_depth, -height / 2])
				rotate([0, -90, 0])
					linear_extrude(height = inner_back_edge_width)
						polygon([
							[-extra, -inner_back_edge_depth],
							[inner_back_edge_height, -inner_back_edge_depth],
							[0, 0],
							[-extra, 0]
							]);

			/* Fillets to round off the edges. */

			union() {

				/* Top left and right rounding. */
			
				translate([-width / 2 + ro, back_depth / 2, height / 2 - ro])
					rotate([0, 0, 180])
						rotate([90, 0, 0])
							fillet(rr, back_depth);
				translate([width / 2 - ro, back_depth / 2, height / 2 - ro])
					rotate([90, 0, 0])
						fillet(rr, back_depth);
			
				/* Top back rounding. */
			
				translate([0, back_depth - ro, height / 2 - ro])
					rotate([0, -90, 0])
						fillet(rr, width);
			
				/* Edge rounding. */
			
				translate([width / 2 - ro, back_depth - ro, 0])
					fillet(rr, height);
				translate([-width / 2 + ro, back_depth - ro, 0])
					rotate([0, 0, 90])
						fillet(rr, height);
			}
		}
}

cartridge();
